@@ -681,6 +681,7 @@ char *prog_name;		/* Name of this program. */
 
 char tc_ent_buf[TC_ENT_SIZE];	/* Buffer for storing a termcap entry. */
 
+/* Modified to work with ncurses termcap emulation - mrn 22 May 96 */
 char *tc_entry(name)
 char *name;			/* Termcap capability string to lookup. */
 {
@@ -689,10 +690,15 @@ char *name;			/* Termcap capability string to lookup. */
  */
 
   char *buf_ptr;
+  char *result;
 
   buf_ptr = tc_ent_buf;
-  if (tgetstr(name, &buf_ptr) == NULL) tc_ent_buf[0] = '\0';
-  return tc_ent_buf;
+  result = tgetstr(name, &buf_ptr);
+  if (result == NULL) {
+    tc_ent_buf[0] = '\0';
+    result = tc_ent_buf;
+  }
+  return result;
 }
 
 void perform_sequence(vcterm)
