@@ -84,7 +84,7 @@ void prbuf (register char *s, register int n);
 #define ESC	'\033'
 #define QUIT	'\034'
 
-struct termio	otty, savetty;
+struct termio	otty, save_tty;
 long		file_pos, file_size;
 int		fnum, no_intty, no_tty, slow_tty;
 int		dum_opt, dlines;
@@ -802,8 +802,16 @@ int getline(register FILE *f, int *length)
 		column = 1 + (column | 7);
 	else if (c == '\b' && column > 0)
 	    column--;
-	else if (c == '\r')
+	else if (c == '\r') {	/* fixes bug with "\r\n" at end of line */
+	    int next = Getc(f);
+            if (next == '\n') {
+              *p--;
+	      Currline++;
+	      break;
+            }
+	    Ungetc(next,f);
 	    column = 0;
+        }
 	else if (c == '\f' && stop_opt) {
 		p[-1] = '^';
 		*p++ = 'L';
@@ -1586,7 +1594,7 @@ retry:
     }
     no_intty = ioctl(fileno(stdin), TCGETA, &otty);
     ioctl(fileno(stderr), TCGETA, &otty);
-    savetty = otty;
+    save_tty = otty;
     slow_tty = (otty.c_cflag & CBAUD) < B1200;
     hardtabs = (otty.c_oflag & TABDLY) != XTABS;
     if (!no_tty) 
@@ -1792,7 +1800,7 @@ void reset_tty ()
 	pstate = 0;
     }
     otty.c_lflag |= ICANON|ECHO;
-    stty(fileno(stderr), &savetty);
+    stty(fileno(stderr), &save_tty);
 }
 
 rdline (f)
